/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "assetmap.h"
#include "asset/asset.h"
#include "assert/assert.h"
#include "asset/assetjson.h"

void assetMapLoad(
  const char_t *path,
  map_t *map
) {
  assertNotNull(map, "assetMapLoad: Map is NULL!");

  // Read in the string data.
  assetOpen(path);
  size_t length = assetGetSize();
  char_t *buffer = malloc(sizeof(char_t) * (length + 1));
  size_t read = assetRead((uint8_t*)buffer, length);
  assertTrue(read == length, "assetMapLoad: Failed to read map data!");
  assetClose();

  // Begin parsing JSON data.
  assetjson_t *json;
  read = assetJsonParse(buffer, &json);
  free(buffer);
  
  assertTrue(
    json->type == ASSET_JSON_DATA_TYPE_OBJECT,
    "assetMapLoad: Map data is not an object!"
  );

  int32_t width = (int32_t)assetJsonGetObjectValue(json, "width")->number;
  int32_t height = (int32_t)assetJsonGetObjectValue(json, "height")->number;

  assetjson_t *layers = assetJsonGetObjectValue(json, "layers");
  assertTrue(
    layers->type == ASSET_JSON_DATA_TYPE_ARRAY,
    "assetMapLoad: Layers is not an array!"
  );

  int32_t layerCount = layers->array.length;
  assertTrue(layerCount == MAP_LAYERS_MAX, "assetMapLoad: No layers found!");

  mapInit(map, width, height, layerCount);

  // Load tile data.
  for(int32_t i = 0; i < layerCount; i++) {
    assetjson_t *layer = layers->array.value[i];
    assertTrue(
      layer->type == ASSET_JSON_DATA_TYPE_OBJECT,
      "assetMapLoad: Layer is not an object!"
    );

    assetjson_t *tiles = assetJsonGetObjectValue(layer, "tiles");
    assertTrue(
      tiles->type == ASSET_JSON_DATA_TYPE_ARRAY,
      "assetMapLoad: Tiles is not an array!"
    );
    assertTrue(
      tiles->array.length == width * height,
      "assetMapLoad: Tile count does not match map size!"
    );

    for(int32_t j = 0; j < width * height; j++) {
      map->tiles[j] = (tile_t){
        .id = (uint16_t)tiles->array.value[j]->number
      };
    }
  }

  // Load entity data
  assetjson_t *entities = assetJsonGetObjectValue(json, "entities");
  assertTrue(
    entities->type == ASSET_JSON_DATA_TYPE_ARRAY,
    "assetMapLoad: Entities is not an array!"
  );

  for(int32_t i = 0; i < entities->array.length; i++) {
    assetjson_t *jEnt = entities->array.value[i];
    assertTrue(
      jEnt->type == ASSET_JSON_DATA_TYPE_OBJECT,
      "assetMapLoad: Entity is not an object!"
    );

    int32_t x = (int32_t)assetJsonGetObjectValue(jEnt, "x")->number;
    int32_t y = (int32_t)assetJsonGetObjectValue(jEnt, "y")->number;
    uint8_t type = (uint8_t)assetJsonGetObjectValue(jEnt, "type")->number;

    entity_t *ent = mapEntityAdd(map);
    entityInit(ent, type, map);
    entityPositionSet(ent, x, y);

    // TODO: Parse any extra data.
  }

  // Load trigger data
  assetjson_t *triggers = assetJsonGetObjectValue(json, "triggers");
  assertTrue(
    triggers->type == ASSET_JSON_DATA_TYPE_ARRAY,
    "assetMapLoad: Triggers is not an array!"
  );

  for(int32_t i = 0; i < triggers->array.length; i++) {
    assetjson_t *jTrig = triggers->array.value[i];
    assertTrue(
      jTrig->type == ASSET_JSON_DATA_TYPE_OBJECT,
      "assetMapLoad: Trigger is not an object!"
    );

    int32_t x = (int32_t)assetJsonGetObjectValue(jTrig, "x")->number;
    int32_t y = (int32_t)assetJsonGetObjectValue(jTrig, "y")->number;
    int32_t width = (int32_t)assetJsonGetObjectValue(jTrig, "width")->number;
    int32_t height = (int32_t)assetJsonGetObjectValue(jTrig, "height")->number;
    uint8_t type = (uint8_t)assetJsonGetObjectValue(jTrig, "type")->number;

    trigger_t *trigger = mapTriggerAdd(map);
    triggerInit(trigger, type, x, y, width, height);
  }

  assetJsonDispose(json);
}